﻿using System;
using System.Collections.Generic;
using UnityEngine;
using AssemblyCSharp;
using System.Collections;

public class BoardManager : MonoBehaviour
{

    public struct Cell
    {
        public Tile tile;
        public GameObject[] highlightLayers;
        public Vertex vertex;
    }

    public int Columns;
    public int Rows;
    public Cell[,] Map;

    public GameObject[] floorTiles;
    public GameObject[] unitSprites;

    public List<Unit> player1Units;
    public List<Unit> player2Units;
    public GameObject cursor;
    public GameObject[] highlightedTiles;

    private Transform boardHolder;

    public AudioClip select;
    public AudioClip cancel;
    public AudioClip die;
    public AudioClip attack;
    public AudioClip cursorMoveSound;

    [HideInInspector] public Unit unitSelected;
    [HideInInspector] private Tile startPosition;
    /*
    String mapFile = @"Assets/Resources/map.txt";
    String enemyUnits = @"Assets/Resources/enemyUnits.txt";
    String allyUnits = @"Assets/Resources/allyUnits.txt";
    */
    public TextAsset mapFile;
    public TextAsset enemyUnits;
    public TextAsset allyUnits;

    [HideInInspector] System.Random random = new System.Random();


    //// Constructor and setup methods 

    private int terrainFactory(Tile t)
    {
        if (t.Terrain.Type == "plain")
            return 0;
        else if (t.Terrain.Type == "forest")
            return 1;
        else if (t.Terrain.Type == "water")
            return 2;
        else if (t.Terrain.Type == "wall")
            return 3;
        else
            return -1;  //Not a valid terrain
    }

    private AssemblyCSharp.Terrain tileFactory(int t)
    {
        if (t == 0) return new Plain();
        else if (t == 1) return new Forest();
        else if (t == 2) return new Water();
        else return null;
    }

    private int unitFactory(Unit u)
    {
        if (u.getClass() == "Cavalier")
        {
            if (u.getOwner() == 1) return 0;
            else if (u.getOwner() == 2) return 1;
            else return -1;
        }
        else if (u.getClass() == "Archer")
        {
            if (u.getOwner() == 1) return 2;
            else if (u.getOwner() == 2) return 3;
            else return -1;
        }
        else if (u.getClass() == "Lancer")
        {
            if (u.getOwner() == 1) return 4;
            else if (u.getOwner() == 2) return 5;
            else return -1;
        }
        else
            return -1;  //Not a valid unit

    }

    private Unit ClassFactory(int t)
    {
        if (t == 1)
        {
            return new Lancer();
        }
        else if (t == 2)
        {
            return new Archer();
        }
        else if (t == 3)
        {
            return new Cavalier();
        }
        else
            return null;  //Not a valid unit
    }




    public void initialiseCursor()
    {
        cursor.transform.position = new Vector3(1, 1, 0);
        cursor = Instantiate(cursor, cursor.transform.position, Quaternion.identity) as GameObject;
        cursor.transform.SetParent(boardHolder);
    }

    public void SetupScene()
    {
        readMapFile();          //prepare map
        readUnitsFiles();        //put units

        initialiseGraph();     // prepare grapg for dijsktra

        terrainSetup();         //animate
        unitSetup();            //animate
    }

    public void readMapFile()
    {
        String[] lines = this.mapFile.text.Split('\n');
        this.Columns = lines[1].Split(' ').Length;
        this.Rows = lines.Length;
        this.Map = new Cell[Columns, Rows];

        for (int y = 0; y < Rows; y++)
        {
            String[] linia = lines[Rows - 1 - y].Split(' ');    // Unity starts the matrix (0,0) at left down instead of left up
            for (int x = 0; x < Columns; x++)
            {
                Map[x, y].tile = new Tile(x, y, tileFactory(int.Parse(linia[x])));

                Map[x, y].highlightLayers = new GameObject[2];
                Map[x, y].highlightLayers[0] = Instantiate(highlightedTiles[0], new Vector2(x, y), Quaternion.identity);
                Map[x, y].highlightLayers[1] = Instantiate(highlightedTiles[1], new Vector2(x, y), Quaternion.identity);
                Map[x, y].highlightLayers[0].SetActive(false);
                Map[x, y].highlightLayers[1].SetActive(false);

                Map[x, y].vertex = new Vertex(x, y);
            }
        }
        Debug.Log("Map reading OK");
    }

    public void readUnitsFiles()
    {
        String[] lines = this.allyUnits.text.Split('\n');

        this.player1Units = new List<Unit>();

        for (int y = 0; y < Rows; y++)
        {
            String[] linia = lines[Rows - 1 - y].Split(' ');    // Unity starts the matrix (0,0) at left down instead of left up
            for (int x = 0; x < Columns; x++)
            {
                if (int.Parse(linia[x]) != 0)
                {
                    Map[x, y].tile.setUnit(ClassFactory(int.Parse(linia[x])));
                    Map[x, y].tile.getUnit().setOwner(1);
                    Map[x, y].tile.getUnit().setPosition(Map[x, y].tile);
                    this.player1Units.Add(Map[x, y].tile.getUnit());
                }
            }
        }
        Debug.Log("Units 1 reading OK");

        lines = this.enemyUnits.text.Split('\n');

        this.player2Units = new List<Unit>();

        for (int y = 0; y < Rows; y++)
        {
            String[] linia = lines[Rows - 1 - y].Split(' ');    // Unity starts the matrix (0,0) at left down instead of left up
            for (int x = 0; x < Columns; x++)
            {
                if (int.Parse(linia[x]) != 0)
                {
                    Map[x, y].tile.setUnit(ClassFactory(int.Parse(linia[x])));
                    Map[x, y].tile.getUnit().setOwner(2);
                    Map[x, y].tile.getUnit().setPosition(Map[x, y].tile);
                    this.player2Units.Add(Map[x, y].tile.getUnit());
                }
            }
        }
        Debug.Log("Units 2 reading OK");

        Debug.Log("player 1 units: " + this.player1Units.Count);
        Debug.Log("player 2 units: " + this.player2Units.Count);
    }

    // Initialise the grid and add the terrain
    void terrainSetup()
    {
        boardHolder = new GameObject("Board").transform;

        for (int y = 0; y < Rows; y++)
        {
            for (int x = 0; x < Columns; x++)
            {
                GameObject toInstantiate = floorTiles[terrainFactory(Map[x, y].tile)];
                GameObject instance = Instantiate(toInstantiate, new Vector2(x, y), Quaternion.identity) as GameObject;
                instance.transform.SetParent(boardHolder);
            }
        }
        Debug.Log("Terrain Setup OK");
    }

    // Put the units in the board
    void unitSetup()
    {
        foreach (Unit u in player1Units)
        {
            GameObject toInstantiate = unitSprites[unitFactory(u)];
            u.setSprite(Instantiate(toInstantiate, new Vector2(u.getPosition().getX(), u.getPosition().getY()), Quaternion.identity) as GameObject);
            u.getSprite().transform.SetParent(boardHolder);
        }

        foreach (Unit u in player2Units)
        {
            GameObject toInstantiate = unitSprites[unitFactory(u)];
            u.setSprite(Instantiate(toInstantiate, new Vector2(u.getPosition().getX(), u.getPosition().getY()), Quaternion.identity) as GameObject);
            u.getSprite().transform.SetParent(boardHolder);
        }
        Debug.Log("Unit Setup OK");
    }


    void initialiseGraph()
    {
        for (int y = 0; y < Rows; y++)
        {
            for (int x = 0; x < Columns; x++)
            {
                addNeighbours(x, y);
            }
        }
        Debug.Log("Graph initialised OK");
    }

    void initialisePath()
    {
        for (int y = 0; y < Rows; y++)
        {
            for (int x = 0; x < Columns; x++)
            {
                Map[x, y].vertex.setMinDist(int.MaxValue);
            }
        }
    }





    //// Public methods

    // move 1 block up down left right each time.
    public void moveCursor(int x, int y)
    {
        Vector3 movement = new Vector2(1 * x, 1 * y);

        if (validPosition(cursor.transform.position + movement))
        {
            cursor.transform.position += movement;
        }
    }

    // user can only move the cursor inside the highlighted area
    public void moveCursorInsideArea(int x, int y)
    {
        Vector3 movement = new Vector2(1 * x, 1 * y);

        if (validPositionInArea(cursor.transform.position + movement))
        {
            cursor.transform.position += movement;
        }
    }

    // Move the cursor to a direct position
    public void moveCursor(Vector2 v)
    {
        if (validPosition(v))
        {
            cursor.transform.position = v;
        }
    }

    public void setUnit(Unit unit, Vector2 pos)
    {
        if (!validPosition(pos))
        {
            Debug.Log("Invalid position");
            return;
        }

        if (this.Map[(int)pos.x, (int)pos.y].tile.Unit == null)
        {
            this.Map[(int)pos.x, (int)pos.y].tile.Unit = unit;
            Debug.Log("Setting the unit " + unit + " in the position " + (int)pos.x + " " + (int)pos.y);
        }
        else
            Debug.Log("Tile " + (int)pos.x + " " + (int)pos.y + " is occupied");
    }


    // unit is the id of the unit; ori i s the initial position and desti is the final position
    public int moveUnit(Unit unit, Vector2 ori, Vector2 dest)
    {
        if (!validPosition(dest))
        {
            Debug.Log("Invalid position");
            return 0;
        }

        // if tile is free or the unit is there, can move
        if (Map[(int)dest.x, (int)dest.y].tile.Unit != null && Map[(int)dest.x, (int)dest.y].tile.Unit != unit)
        {
            Debug.Log("Tile " + dest.x + " " + dest.y + " is occupied");
            return 0;
        }
        else
        {
            this.Map[(int)ori.x, (int)ori.y].tile.Unit = null;
            this.Map[(int)dest.x, (int)dest.y].tile.Unit = unit;
            this.Map[(int)dest.x, (int)dest.y].tile.Unit.move(Map[(int)dest.x, (int)dest.y].tile);
            Debug.Log("unit " + unit + " in the position " + (int)ori.x + " " + (int)ori.y + " moved to the position " + (int)dest.x + " " + (int)dest.y);
            return 1;
        }
    }

    // must have a unit selected. Return 1 if move has been done; else 0
    public int moveSelectedUnit()
    {
        return moveUnit(unitSelected, new Vector2(unitSelected.getPosition().getX(), unitSelected.getPosition().getY()), new Vector2((int)cursor.transform.position.x, (int)cursor.transform.position.y));
    }

    public void returnSelectedUnitToStartingTile()
    {
        SoundManager.instance.PlaySfx(this.cancel);
        moveUnit(unitSelected, new Vector2(unitSelected.getPosition().getX(), unitSelected.getPosition().getY()), new Vector2(startPosition.getX(), startPosition.getY()));
        moveCursor(new Vector2(startPosition.getX(), startPosition.getY()));
    }

    // action domain is {"move", "attack"}
    public int selectTile(String action)
    {
        Tile focus = Map[(int)cursor.transform.position.x, (int)cursor.transform.position.y].tile;

        if (focus.Unit != null)
        {
            //ally unit
            if (focus.Unit.getOwner() == 1 && focus.Unit.canAct())
            {
                paintHighlightedTiles(focus.Unit, action);
                if (this.unitSelected == null)
                {
                    SoundManager.instance.PlaySfx(this.select);
                    this.unitSelected = focus.Unit;
                    this.startPosition = focus;
                }
                return 1;
            }
        }
        return 0;
    }

    public int attackOrWait()
    {
        Tile focus = Map[(int)cursor.transform.position.x, (int)cursor.transform.position.y].tile;

        // attack enemy unit
        if (focus.Unit != null && focus.Unit.getOwner() != this.unitSelected.getOwner())
        {
            // this.unitSelected attack focus.Unit
            combat(focus.Unit);
            return 1;
        }
        // wait
        else if (focus.Unit != null && focus.Unit == this.unitSelected)
        {
            return 1;
        }
        return 0;
    }

    public void combat(Unit enemy)
    {
        // unitSelected attacks enemy
        SoundManager.instance.PlaySfx(this.attack);
        enemy.loseHP(trueDamage(enemy));

        if (enemy.isDead())
        {
            deleteUnitFromPlayer(enemy);
        }
        else if (canCounterattack(enemy))
        {
            SoundManager.instance.PlaySfx(this.attack);
            this.unitSelected.loseHP(trueCounterattack(enemy));
            if (this.unitSelected.isDead())
            {
                deleteUnitFromPlayer(this.unitSelected);
            }
        }
    }

    public void freeUnitSelected()
    {
        SoundManager.instance.PlaySfx(this.cancel);
        this.unitSelected.idle();
        this.unitSelected = null;
        this.startPosition = null;
    }

    public void endUnitTurn()
    {
        this.unitSelected.wait();
        this.unitSelected = null;
        this.startPosition = null;
    }

    public void enemyTurn()
    {
        // posem com a selected la unit actual
        this.unitSelected = nextUnitAvailable(2);
        this.startPosition = unitSelected.getPosition();

        moveCursor(new Vector2(this.startPosition.getX(), this.startPosition.getY()));
        System.Threading.Thread.Sleep(500);

        List<Cell> movementZone = getCellsInRange(this.unitSelected, "move");   //llista de posicions on podem anar
        HashSet<Cell> attackZone = enemiesInRange(movementZone);                //llista d'enemics dintre el rang

        if (attackZone.Count != 0)
        {
            //enemies in range
            Unit victim = evaluateEnemiesAndReturnBest(attackZone);
            Tile bestSpot = searchBestCellToAttack(victim.getPosition(), movementZone);
            moveCursor(new Vector2(bestSpot.getX(), bestSpot.getY()));
            System.Threading.Thread.Sleep(500);
            moveUnit(this.unitSelected, new Vector2(this.startPosition.getX(), this.startPosition.getY()), new Vector2(bestSpot.getX(), bestSpot.getY()));
            combat(victim);
        }
        else
        {
            moveUnitToRandomPosition(movementZone);
        }

        endUnitTurn();
    }

    public void startTurn(int player)
    {

        if (player == 1)
        {
            foreach (Unit u in player1Units)
            {
                u.idle();
            }
        }

        else if (player == 2)
        {


            foreach (Unit u in player2Units)
            {
                u.idle();
            }
        }
    }

    public Unit nextUnitAvailable(int player)
    {
        if (player == 1)
        {
            return player1Units.Find(x => x.canAct());
        }
        else if (player == 2)
        {
            return player2Units.Find(x => x.canAct());
        }
        else
        {
            Debug.Log("Unknown player");
        }
        return null;
    }

    public void clearHighlightedTiles()
    {
        for (int y = 0; y < Rows; y++)
        {
            for (int x = 0; x < Columns; x++)
            {
                Map[x, y].highlightLayers[1].SetActive(false);
                Map[x, y].highlightLayers[0].SetActive(false);
            }
        }
    }

    public void endPlayerTurn()
    {
        foreach (Unit u in player1Units)
        {
            u.wait();
        }
    }



    //// Private methods

    private void moveUnitToRandomPosition(List<Cell> list)
    {

        int r = this.random.Next(list.Count);
        int tries = 3;
        while (list[r].tile.getUnit() != null && tries > 0)
        {
            r = this.random.Next(list.Count);
            tries--;
        }

        System.Threading.Thread.Sleep(500);
        if (tries == 0)
        {
            moveCursor(new Vector2(this.startPosition.getX(), this.startPosition.getY()));
            moveUnit(this.unitSelected, new Vector2(this.startPosition.getX(), this.startPosition.getY()), new Vector2(this.startPosition.getX(), this.startPosition.getY()));
        }
        else
        {
            moveCursor(new Vector2(list[r].tile.getX(), list[r].tile.getY()));
            moveUnit(this.unitSelected, new Vector2(this.startPosition.getX(), this.startPosition.getY()), new Vector2(list[r].tile.getX(), list[r].tile.getY()));
        }
    }

    private Unit evaluateEnemiesAndReturnBest(HashSet<Cell> set)
    {
        Unit victim = null;
        int dmg = 0;
        int ca = 10;

        foreach (Cell c in set)
        {
            if (enemyDies(c.tile.getUnit()))
            {
                return c.tile.getUnit();
            }
            else if (dmg < trueDamage(c.tile.getUnit()))
            {
                victim = c.tile.getUnit();
                dmg = trueDamage(c.tile.getUnit());
                ca = trueCounterattack(c.tile.getUnit());
            }
            else if (dmg == trueDamage(c.tile.getUnit()) && ca > trueCounterattack(c.tile.getUnit()))
            {
                victim = c.tile.getUnit();
                ca = trueCounterattack(c.tile.getUnit());
            }
        }

        return victim;
    }

    private Tile searchBestCellToAttack(Tile objective, List<Cell> zone)
    {
        List<Cell> candidates = new List<Cell>();

        List<Cell> attackRange = getCellsInAttackRange(this.unitSelected.getRange(), objective);

        // search for candidates
        foreach (Cell c in zone)
        {
            if (attackRange.Contains(c) && c.tile.getUnit() == null)
            {
                candidates.Add(c);
            }
        }

        return bestPosition(candidates);
    }

    private Tile bestPosition(List<Cell> list)
    {
        Tile bestSpot = null;
        int def = -2;

        foreach (Cell c in list)
        {
            if (c.tile.getTerrainDef() == 1)
            {
                return c.tile;
            }
            else if (c.tile.getTerrainDef() > def)
            {
                bestSpot = c.tile;
                def = c.tile.getTerrainDef();
            }
        }

        return bestSpot;
    }

    private bool enemyDies(Unit u)
    {
        return ((this.unitSelected.damageDealtTo(u.getClass()) + u.getPosition().getTerrainDef()) >= u.getCurrentHP());
    }

    private int trueDamage(Unit u)
    {
        return this.unitSelected.damageDealtTo(u.getClass()) - u.getPosition().getTerrainDef();
    }

    private int trueCounterattack(Unit u)
    {
        return u.damageDealtTo(this.unitSelected.getClass()) - this.unitSelected.getPosition().getTerrainDef();
    }

    private HashSet<Cell> enemiesInRange(List<Cell> movement)
    {
        List<Cell> aux;
        HashSet<Cell> enemiesInRange = new HashSet<Cell>();
        foreach (Cell cell in movement)
        {
            if (cell.tile.getUnit() == null)    // if this unit can move there
            {
                aux = getCellsInAttackRange(this.unitSelected.getRange(), cell.tile);
                foreach (Cell possibleEnemy in aux)
                {
                    //if we find an enemy, we add it to the list
                    if (possibleEnemy.tile.getUnit() != null && possibleEnemy.tile.getUnit().getOwner() != 2)
                    {
                        enemiesInRange.Add(possibleEnemy);
                    }
                }
            }
        }

        return enemiesInRange;
    }

    private void addNeighbours(int x, int y)
    {
        if (validPosition(new Vector2(x - 1, y)))
        {
            this.Map[x, y].vertex.addRelation(new Relation(this.Map[x - 1, y].vertex, this.Map[x - 1, y].tile.getMoveRed()));
        }
        if (validPosition(new Vector2(x + 1, y)))
        {
            this.Map[x, y].vertex.addRelation(new Relation(this.Map[x + 1, y].vertex, this.Map[x + 1, y].tile.getMoveRed()));
        }
        if (validPosition(new Vector2(x, y + 1)))
        {
            this.Map[x, y].vertex.addRelation(new Relation(this.Map[x, y + 1].vertex, this.Map[x, y + 1].tile.getMoveRed()));
        }
        if (validPosition(new Vector2(x, y - 1)))
        {
            this.Map[x, y].vertex.addRelation(new Relation(this.Map[x, y - 1].vertex, this.Map[x, y - 1].tile.getMoveRed()));
        }
    }

    // Paint the cell list obtained by the unit u and the option s
    private void paintHighlightedTiles(Unit u, String s)
    {
        List<Cell> positions = getCellsInRange(u, s);

        int type;
        if (s == "move") { type = 0; u.move(); }
        else if (s == "attack") { type = 1; u.idle(); }
        else { Debug.Log("Not a valid type"); return; }

        foreach (Cell cell in positions)
        {
            Map[cell.tile.getX(), cell.tile.getY()].highlightLayers[type].SetActive(true);
        }
    }

    // Check if position v is in the limits
    private bool validPosition(Vector2 v)
    {
        return (v.x >= 0) && (v.x < Columns) && (v.y >= 0) && (v.y < Rows);
    }

    private bool validPositionInArea(Vector2 v)
    {
        return (v.x >= 0) && (v.x < Columns) && (v.y >= 0) && (v.y < Rows) && (Map[(int)v.x, (int)v.y].highlightLayers[1].activeSelf || Map[(int)v.x, (int)v.y].highlightLayers[0].activeSelf);
    }

    // Return a list containing all the cells in range of the unit and the action 
    private List<Cell> getCellsInRange(Unit u, String action)
    {
        int range = 0;
        List<Cell> positions = new List<Cell>();
        if (action == "attack")
        {
            range = u.getRange();
            positions = getCellsInAttackRange(range, u.getPosition());
        }
        else if (action == "move")
        {
            range = u.getMovement();
            positions = getCellsInMovementRange(range, u.getPosition());
        }
        else
        {
            Debug.Log("Not a valid action");
        }

        return positions;

    }

    // Return a list containg all the cells in a radius of range and pos in the center
    private List<Cell> getCellsInAttackRange(int range, Tile pos)
    {
        List<Cell> positions = new List<Cell>();
        for (int i = 0; i <= range; i++)
        {
            if ((pos.getY() - range + i) >= 0)
            {
                positions.Add(Map[pos.getX(), pos.getY() - range + i]);
                for (int j = 1; j <= i; j++)
                {
                    if ((pos.getX() + j) < this.Columns) positions.Add(Map[pos.getX() + j, pos.getY() - range + i]);
                    if ((pos.getX() - j) >= 0) positions.Add(Map[pos.getX() - j, pos.getY() - range + i]);
                }
            }

            if ((pos.getY() + range - i) < this.Rows && i < range)
            {
                positions.Add(Map[pos.getX(), pos.getY() + range - i]);
                for (int j = 1; j <= i; j++)
                {
                    if ((pos.getX() + j) < this.Columns) positions.Add(Map[pos.getX() + j, pos.getY() + range - i]);
                    if ((pos.getX() - j) >= 0) positions.Add(Map[pos.getX() - j, pos.getY() + range - i]);
                }
            }
        }

        return positions;
    }

    // Return a list containg the cells accessible by pos
    private List<Cell> getCellsInMovementRange(int range, Tile pos)
    {
        List<Cell> positions = new List<Cell>();
        LinkedList<Vertex> list = new LinkedList<Vertex>();

        initialisePath();
        Map[pos.getX(), pos.getY()].vertex.setMinDist(0);
        list.AddFirst(Map[pos.getX(), pos.getY()].vertex);
        positions.Add(Map[pos.getX(), pos.getY()]);


        while (list.First != null)
        {
            Vertex act = list.First.Value;
            list.RemoveFirst();

            foreach (Relation r in act.getNearBy())
            {
                if (isBetterWay(act, r, range))
                {
                    r.getVertex().setMinDist(act.getMinDist() + r.getCost());
                    r.getVertex().setPrevious(act);
                    list.AddLast(r.getVertex());
                    positions.Add(Map[r.getVertex().getX(), r.getVertex().getY()]);
                }
            }
        }

        return positions;
    }

    // Evaluate if the path start->nextTo is cheaper than the previous value
    private bool isBetterWay(Vertex start, Relation nextTo, int limit)
    {

        int oldPathCost = nextTo.getVertex().getMinDist();
        int newPathCost = start.getMinDist() + nextTo.getCost();
        return (newPathCost <= limit && newPathCost < oldPathCost);
    }

    private void deleteUnitFromPlayer(Unit u)
    {
        if (u.getOwner() == 1)
        {
            SoundManager.instance.PlaySfx(this.die);
            this.Map[u.getPosition().getX(), u.getPosition().getY()].tile.Unit = null;
            this.player1Units.Remove(u);
            Destroy(u.getSprite());
            u = null;
        }
        else if (u.getOwner() == 2)
        {
            SoundManager.instance.PlaySfx(this.die);
            this.Map[u.getPosition().getX(), u.getPosition().getY()].tile.Unit = null;
            this.player2Units.Remove(u);
            Destroy(u.getSprite());
            u = null;
        }
        else
        {
            Debug.Log("Unknown player");
        }
    }

    // Return true if unitSelected is in range of enemy
    private bool canCounterattack(Unit enemy)
    {
        List<Cell> cellsInRange = getCellsInRange(enemy, "attack");

        return cellsInRange.Contains(Map[this.unitSelected.getPosition().getX(), this.unitSelected.getPosition().getY()]);
    }

}
