﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.SceneManagement;

public class GameManager : MonoBehaviour
{

    public BoardManager boardScript;
    [HideInInspector] public bool playersTurn;
    [HideInInspector] string stage = "defaultMap";
    [HideInInspector] bool victory;
    [HideInInspector] bool stillPlaying = true;
    public float time = 1.0f;

    public AudioClip playerTurnMusic;
    public AudioClip enemyTurnMusic;
    public AudioClip victoryMusic;
    public AudioClip gameOverMusic;
    public AudioClip mainMenuMusic;


    // Use this for initialization
    void Start()
    {
        DontDestroyOnLoad(gameObject);
        boardScript = GetComponent<BoardManager>();
        InitGame();
        SoundManager.instance.PlayMusic(this.playerTurnMusic);
    }

    void InitGame()
    {
        boardScript.SetupScene();
        boardScript.initialiseCursor();
    }

    void EndGame()
    {
        if (!this.stillPlaying && this.victory)
        {
            Debug.Log("victory");
            SoundManager.instance.PlayMusic(this.victoryMusic);
            SceneManager.LoadScene(3);
        }
        else if (!this.stillPlaying && !this.victory)
        {
            Debug.Log("defeat");
            SoundManager.instance.PlayMusic(this.gameOverMusic);
            SceneManager.LoadScene(2);
        }
    }

    // Update is called once per frame
    void Update()
    {
        if (boardScript.player1Units.Count == 0)
        {
            this.stillPlaying = false;
            this.victory = false;
            enabled = false;
            EndGame();
        }

        else if (boardScript.player2Units.Count == 0)
        {
            this.stillPlaying = false;
            this.victory = true;
            enabled = false;
            EndGame();
        }

        else if (this.playersTurn && this.stillPlaying)
        {
            time++;
            readInput();
            if (boardScript.nextUnitAvailable(1) == null)
            {
                this.playersTurn = false;
                boardScript.startTurn(2);
                SoundManager.instance.PlayMusic(this.enemyTurnMusic);
            }
        }

        else if (!this.playersTurn && this.stillPlaying)
        {
            boardScript.enemyTurn();
            if (boardScript.nextUnitAvailable(2) == null)
            {
                this.playersTurn = true;
                boardScript.startTurn(1);
                SoundManager.instance.PlayMusic(this.playerTurnMusic);
            }

        }
    }

    void readInput()
    {
        if (stage == "defaultMap") { readInputInIdleMap(); }
        else if (stage == "moveUnitMap") { readInputMoveUnit(); }
        else if (stage == "attackUnitMap") { readInputAttackUnit(); }
    }

    // when nothing is selected
    void readInputInIdleMap()
    {
        if (Input.GetKeyDown(KeyCode.A))
        {
            int result = boardScript.selectTile("move");
            if (result == 1) { stage = "moveUnitMap"; }
            // saber si seleccionem una unit, edifici o res. En funció d'això, modificar stage
        }

        if (Input.GetKeyDown(KeyCode.Backspace))
        {
            boardScript.endPlayerTurn();
        }

        else if (time > 4)
        {
            if (Input.GetKey(KeyCode.UpArrow))
                boardScript.moveCursor(0, 1);
            else if (Input.GetKey(KeyCode.DownArrow))
                boardScript.moveCursor(0, -1);
            else if (Input.GetKey(KeyCode.LeftArrow))
                boardScript.moveCursor(-1, 0);
            else if (Input.GetKey(KeyCode.RightArrow))
                boardScript.moveCursor(1, 0);

            time = 0;
        }
    }

    //when a unit is selected
    void readInputMoveUnit()
    {
        if (Input.GetKeyDown(KeyCode.A))
        {
            int result = boardScript.moveSelectedUnit();
            if (result == 1)
            {
                boardScript.clearHighlightedTiles();
                boardScript.selectTile("attack");
                stage = "attackUnitMap";
            }
        }
        else if (Input.GetKeyDown(KeyCode.S))
        {
            //cancel move attempt and de-seelct unit
            boardScript.clearHighlightedTiles();
            boardScript.freeUnitSelected();
            stage = "defaultMap";
        }

        else if (time > 4)
        {
            if (Input.GetKey(KeyCode.UpArrow))
                boardScript.moveCursorInsideArea(0, 1);
            else if (Input.GetKey(KeyCode.DownArrow))
                boardScript.moveCursorInsideArea(0, -1);
            else if (Input.GetKey(KeyCode.LeftArrow))
                boardScript.moveCursorInsideArea(-1, 0);
            else if (Input.GetKey(KeyCode.RightArrow))
                boardScript.moveCursorInsideArea(1, 0);

            time = 0;
        }
    }

    void readInputAttackUnit()
    {
        if (Input.GetKeyDown(KeyCode.A))
        {
            int result = boardScript.attackOrWait();
            if (result == 1)
            {
                boardScript.clearHighlightedTiles();
                boardScript.endUnitTurn();
                stage = "defaultMap";
            }
        }
        else if (Input.GetKeyDown(KeyCode.S))
        {
            Debug.Log("cancel");
            boardScript.clearHighlightedTiles();
            boardScript.returnSelectedUnitToStartingTile();
            boardScript.selectTile("move");
            stage = "moveUnitMap";
        }

        else if (time > 4)
        {
            if (Input.GetKey(KeyCode.UpArrow))
                boardScript.moveCursorInsideArea(0, 1);
            else if (Input.GetKey(KeyCode.DownArrow))
                boardScript.moveCursorInsideArea(0, -1);
            else if (Input.GetKey(KeyCode.LeftArrow))
                boardScript.moveCursorInsideArea(-1, 0);
            else if (Input.GetKey(KeyCode.RightArrow))
                boardScript.moveCursorInsideArea(1, 0);

            time = 0;
        }
    }
}
