﻿using System;
using System.Collections.Generic;
using System.Threading;
using UnityEngine;

namespace AssemblyCSharp
{
    public abstract class Unit
    {
        static int NextId;
        GameObject sprite;
        public int Id { get; }
        private String Class { get; }
        private Tile Position { get; set; }
        private int maxHP { get; set; }
        private int currentHP { get; set; }
        private int ATK { get; set; }
        private int DEF { get; set; }
        private int MOV { get; set; }
        private int RNG { get; set; }
        private int Owner;
        private List<String> Advantage { get; }

        public Unit() { }

        public Unit(String c, int hp, int atk, int def, int mov, int r)
        {
            this.Id = Interlocked.Increment(ref NextId);
            this.Class = c;
            this.maxHP = hp;
            this.currentHP = hp;
            this.ATK = atk;
            this.DEF = def;
            this.MOV = mov;
            this.RNG = r;
            this.Advantage = GetAdvantages(c);

            //must be set later
            this.Owner = 0;
            this.sprite = null;
        }

        public int getId()
        {
            return this.Id;
        }

        public String getClass()
        {
            return this.Class;
        }

        public int getRange()
        {
            return this.RNG;
        }

        public Tile getPosition()
        {
            return this.Position;
        }

        public void setPosition(Tile p)
        {
            this.Position = p;
        }

        public int getMovement()
        {
            return this.MOV;
        }

        public int getOwner()
        {
            return this.Owner;
        }

        public void setOwner(int o)
        {
            this.Owner = o;
        }

        public void recoverHP()
        {
            this.currentHP = Math.Min(this.currentHP + this.maxHP * 2 / 10, this.maxHP);
            Debug.Log("Unit " + Id + " recovered " + this.maxHP * 2 / 10 + " HP and now it has " + this.currentHP);
        }

        public void loseHP( int dmg)
        {
            this.currentHP -= dmg;
        }

        public int getCurrentHP()
        {
            return this.currentHP;
        }

        public int getAtk()
        {
            return this.ATK;
        }

        public int getDef()
        {
            return this.DEF;
        }

        public bool isDead()
        {
            return this.currentHP < 1;
        }

        public void move(Tile t)
        {
            this.Position = t;
            this.sprite.transform.position = new Vector2(t.getX(), t.getY());
            this.sprite.GetComponent<Animator>().SetBool("move", false);
        }

        public void setSprite(GameObject s)
        {
            this.sprite = s;
            this.sprite.GetComponent<Animator>().SetBool("move", false);
            this.sprite.GetComponent<Animator>().SetBool("wait", false);
        }

        public GameObject getSprite()
        {
            return this.sprite;
        }

        public void wait()
        {
            this.sprite.GetComponent<Animator>().SetBool("wait", true);
        }

        public bool canAct()
        {
            return !this.sprite.GetComponent<Animator>().GetBool("wait");
        }

        public void move()
        {
            this.sprite.GetComponent<Animator>().SetBool("move", true);
        }

        public void idle()
        {
            this.sprite.GetComponent<Animator>().SetBool("move", false);
            this.sprite.GetComponent<Animator>().SetBool("wait", false);
        }

        public int damageDealtTo(String enemy)
        {
            if (this.Advantage.Contains(enemy)) { return this.ATK * 2; }
            else { return this.ATK; }
        }



        private List<String> GetAdvantages(String c)
        {
            switch (c)
            {
                case "Cavalier":
                    return Cavalier;
                case "Archer":
                    return Archer;
                case "Lancer":
                    return Lancer;
                default:
                    return null;
            }
        }

        static List<String> Cavalier = new List<String> { "Archer" };
        static List<String> Lancer = new List<String> { "Cavalier" };
        static List<String> Archer = new List<String> { "Lancer" };
    }
}